From 0af05ac01be853340b3b53862de9853d8d44c0c6 Mon Sep 17 00:00:00 2001
From: liuyingdong <liuyingdong@huawei.com>
Date: Thu, 2 Feb 2017 19:40:48 +0200
Subject: introduce suspend_cancel mechanism for frontend devices.
1.1 On a cancelled suspend, xen frontend devices need know their state is invariant.
1.2 On a cancelled suspend the vcpu_info location does not change 
(it's still in the per-cpu area registered by xen_hvm_cpu_init()). 
So do not call xen_hvm_init_shared_info_page() which would make the kernel think its back in the shared info. 
With the wrong vcpu_info, events cannot be received and the domain will hang after a cancelled suspend.

---
 dev/xen/blkfront/blkfront.c |  5 +++++
 dev/xen/control/control.c   | 11 +++++++++--
 dev/xen/netfront/netfront.c | 21 +++++++++++++++++++++
 x86/xen/hvm.c               | 16 ++++++++++------
 xen/xenbus/xenbusb.c        | 35 +++++++++++++++++++----------------
 xen/xenbus/xenbusvar.h      |  2 ++
 6 files changed, 66 insertions(+), 24 deletions(-)

diff --git a/dev/xen/blkfront/blkfront.c b/dev/xen/blkfront/blkfront.c
index 9eca220..47cd83f 100644
--- a/dev/xen/blkfront/blkfront.c
+++ b/dev/xen/blkfront/blkfront.c
@@ -1537,6 +1537,11 @@ xbd_resume(device_t dev)
 {
 	struct xbd_softc *sc = device_get_softc(dev);
 
+	if(g_suspend_cancelled) {
+		sc->xbd_state = XBD_STATE_CONNECTED;
+		return (0);
+	}
+
 	DPRINTK("xbd_resume: %s\n", xenbus_get_node(dev));
 
 	xbd_free(sc);
diff --git a/dev/xen/control/control.c b/dev/xen/control/control.c
index 58fefcc..4f8471f 100644
--- a/dev/xen/control/control.c
+++ b/dev/xen/control/control.c
@@ -190,6 +190,11 @@ xctrl_reboot()
 	shutdown_nice(0);
 }
 
+static void _set_suspend_cancelled(bool suspend_cancelled)
+{
+	g_suspend_cancelled = suspend_cancelled;
+}
+
 static void
 xctrl_suspend()
 {
@@ -278,7 +283,9 @@ xctrl_suspend()
 	/*
 	 * Reset grant table info.
 	 */
-	gnttab_resume(NULL);
+	if(suspend_cancelled == 0)  {
+		gnttab_resume(NULL);
+	}
 
 #ifdef SMP
 	if (!CPU_EMPTY(&cpu_suspend_map)) {
@@ -296,6 +303,7 @@ xctrl_suspend()
 	 * FreeBSD really needs to add DEVICE_SUSPEND_CANCEL or
 	 * similar.
 	 */
+	_set_suspend_cancelled(suspend_cancelled != 0);
 	DEVICE_RESUME(root_bus);
 	mtx_unlock(&Giant);
 
@@ -319,7 +327,6 @@ xctrl_suspend()
 #endif
 
 	resume_all_proc();
-
 	EVENTHANDLER_INVOKE(power_resume);
 
 	if (bootverbose)
diff --git a/dev/xen/netfront/netfront.c b/dev/xen/netfront/netfront.c
index 459712a..537018a 100644
--- a/dev/xen/netfront/netfront.c
+++ b/dev/xen/netfront/netfront.c
@@ -153,6 +153,10 @@ static int xn_get_responses(struct netfront_rxq *,
     struct netfront_rx_info *, RING_IDX, RING_IDX *,
     struct mbuf **);
 
+#ifdef INET
+static void netfront_send_fake_arp(device_t dev, struct netfront_info *info);
+#endif
+
 #define virt_to_mfn(x) (vtophys(x) >> PAGE_SHIFT)
 
 #define INVALID_P2M_ENTRY (~0UL)
@@ -440,6 +444,23 @@ netfront_resume(device_t dev)
 {
 	struct netfront_info *info = device_get_softc(dev);
 
+	if(g_suspend_cancelled) {
+		u_int i;
+		for (i = 0; i < info->num_queues; i++) {
+		    XN_RX_LOCK(&info->rxq[i]);
+		    XN_TX_LOCK(&info->txq[i]);
+		}
+		netfront_carrier_on(info);
+		for (i = 0; i < info->num_queues; i++) {
+		    XN_RX_UNLOCK(&info->rxq[i]);
+		    XN_TX_UNLOCK(&info->txq[i]);
+		}
+#ifdef INET
+		netfront_send_fake_arp(dev, info);
+#endif
+		return (0);
+	}
+
 	netif_disconnect_backend(info);
 	return (0);
 }
diff --git a/x86/xen/hvm.c b/x86/xen/hvm.c
index e10659e..c96b7be 100644
--- a/x86/xen/hvm.c
+++ b/x86/xen/hvm.c
@@ -297,10 +297,9 @@ xen_hvm_init(enum xen_hvm_init_type init_type)
 	int error;
 	int i;
 
-	if (init_type == XEN_HVM_INIT_CANCELLED_SUSPEND)
-		return;
-
-	error = xen_hvm_init_hypercall_stubs(init_type);
+	if (init_type != XEN_HVM_INIT_CANCELLED_SUSPEND) {
+		error = xen_hvm_init_hypercall_stubs(init_type);
+	}
 
 	switch (init_type) {
 	case XEN_HVM_INIT_COLD:
@@ -331,6 +330,8 @@ xen_hvm_init(enum xen_hvm_init_type init_type)
 		CPU_FOREACH(i)
 			DPCPU_ID_SET(i, vcpu_info, NULL);
 		break;
+	case XEN_HVM_INIT_CANCELLED_SUSPEND:
+		break;
 	default:
 		panic("Unsupported HVM initialization type");
 	}
@@ -344,7 +345,9 @@ xen_hvm_init(enum xen_hvm_init_type init_type)
 	 * is passed inside the start_info struct and is already set, so this
 	 * functions are no-ops.
 	 */
-	xen_hvm_init_shared_info_page();
+	if (init_type != XEN_HVM_INIT_CANCELLED_SUSPEND) {
+		xen_hvm_init_shared_info_page();
+	}
 	xen_hvm_disable_emulated_devices();
 } 
 
@@ -361,7 +364,8 @@ xen_hvm_resume(bool suspend_cancelled)
 	    XEN_HVM_INIT_CANCELLED_SUSPEND : XEN_HVM_INIT_RESUME);
 
 	/* Register vcpu_info area for CPU#0. */
-	xen_hvm_cpu_init();
+	if(!suspend_cancelled)
+		xen_hvm_cpu_init();
 }
  
 static void
diff --git a/xen/xenbus/xenbusb.c b/xen/xenbus/xenbusb.c
index 3a0e543..fa9ba61 100644
--- a/xen/xenbus/xenbusb.c
+++ b/xen/xenbus/xenbusb.c
@@ -791,29 +791,32 @@ xenbusb_resume(device_t dev)
 			if (device_get_state(kids[i]) == DS_NOTPRESENT)
 				continue;
 
-			ivars = device_get_ivars(kids[i]);
+			if(!g_suspend_cancelled) {
+				ivars = device_get_ivars(kids[i]);
 
-			xs_unregister_watch(&ivars->xd_otherend_watch);
-			xenbus_set_state(kids[i], XenbusStateInitialising);
+				xs_unregister_watch(&ivars->xd_otherend_watch);
+				xenbus_set_state(kids[i], XenbusStateInitialising);
 
-			/*
-			 * Find the new backend details and
-			 * re-register our watch.
-			 */
-			error = XENBUSB_GET_OTHEREND_NODE(dev, ivars);
-			if (error)
-				return (error);
+				/*
+				 * Find the new backend details and
+				 * re-register our watch.
+				 */
+				error = XENBUSB_GET_OTHEREND_NODE(dev, ivars);
+				if (error)
+					return (error);
 
-			statepath = malloc(ivars->xd_otherend_path_len
-			    + strlen("/state") + 1, M_XENBUS, M_WAITOK);
-			sprintf(statepath, "%s/state", ivars->xd_otherend_path);
+				statepath = malloc(ivars->xd_otherend_path_len
+				    + strlen("/state") + 1, M_XENBUS, M_WAITOK);
+				sprintf(statepath, "%s/state", ivars->xd_otherend_path);
 
-			free(ivars->xd_otherend_watch.node, M_XENBUS);
-			ivars->xd_otherend_watch.node = statepath;
+				free(ivars->xd_otherend_watch.node, M_XENBUS);
+				ivars->xd_otherend_watch.node = statepath;
+			}
 
 			DEVICE_RESUME(kids[i]);
 
-			xs_register_watch(&ivars->xd_otherend_watch);
+			if(!g_suspend_cancelled)
+				xs_register_watch(&ivars->xd_otherend_watch);
 #if 0
 			/*
 			 * Can't do this yet since we are running in
diff --git a/xen/xenbus/xenbusvar.h b/xen/xenbus/xenbusvar.h
index 377d60c..02a5bfa 100644
--- a/xen/xenbus/xenbusvar.h
+++ b/xen/xenbus/xenbusvar.h
@@ -99,6 +99,8 @@ XENBUS_ACCESSOR(state,		STATE,			enum xenbus_state)
 XENBUS_ACCESSOR(otherend_id,	OTHEREND_ID,		int)
 XENBUS_ACCESSOR(otherend_path,	OTHEREND_PATH,		const char *)
 
+bool g_suspend_cancelled;
+
 /**
  * Return the state of a XenBus device.
  *
-- 
2.11.0.windows.3

